import _io
import torch
import pickle
from torchvision import transforms
from PIL import Image

from ocrd_typegroups_classifier.data.classmap import ClassMap
from ocrd_typegroups_classifier.data.classmap import IndexRemap



class TypegroupsClassifier:
    """ Class wrapping type group information and a classifier.
    
        Attributes
        ----------
        
        classMap: ClassMap
            Maps class names to indices corresponding to what the network
            outputs.
        network: PyTorch network
            Classifier
        dev: str
            Device on which the data must be processed
    
    """
    
    def __init__(self, groups, network, device=None):
        """ Constructor of the class.
        
            Parameters
            ----------
            
            groups: map string to int
                Maps names to IDs with regard to the network outputs;
                note that several names can point to the same ID, but
                the inverse is not possible.
            network:PyTorch network
                This network has to have the same interface as the
                VRAEC, return three values when being called: the
                classification result, a variational loss, and a feature
                vector.
            device: str
                Device on which the data has to be processed; if not set,
                then either the cpu or cuda:0 will be used.
        
        """
        
        self.classMap = ClassMap(groups)
        self.network = network
        if device is None:
            self.dev = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        else:
            self.dev = device
        network.to(self.dev)
    
    @classmethod
    def load(cls, input):
        """ Loads a type groups classifier from a file
            
            Parameters
            ----------
            input: string or file
                File or path to the file from which the instance has to
                be loaded.
        
        """
        
        if type(input) is str:
            f = open(input, 'rb')
            res = cls.load(f)
            f.close()
            return res
        if not type(input) is _io.BufferedReader:
            raise Exception('TypegroupsClassifier.load() requires a string or a file')
        res = pickle.load(input)
        # If trained with CUDA and loaded on a device without CUDA
        res.dev = torch.device(res.dev if torch.cuda.is_available() else "cpu")
        res.network.to(res.dev)
        return res
        
    def save(self, output):
        """ Stores the instance to a file
        
            Parameters
            ----------
                output: string or file
                    File or path to the file to which the instane has to
                    be stored.
        """
        
        if type(output) is str:
            f = open(output, 'wb')
            self.save(f)
            f.close()
            return
        if not type(output) is _io.BufferedWriter:
            raise Exception('save() requires a string or a file')
        # Moving the network to the cpu so that it can be reloaded on
        # machines which do not have CUDA available.
        self.network.to("cpu")
        pickle.dump(self, output)
        self.network.to(self.dev)
    
    def filter(self, sample, label):
        """ Removes data with unknown type groups
            
            Parameters
            ----------
                sample: PyTorch tensor
                    Tensor of inputs for the network
                label: PyTorch tensor
                    Tensor of class IDs, the unknown ones being set to -1
            
            Returns
            -------
                sample, label
                    The input tensors without the ones having a -1 label
        """
        
        selection = label!=-1
        return sample[selection], label[selection]
    
    def run(self, pil_image, stride, batch_size=32, score_as_key=False):
        return self.classify(pil_image, stride, batch_size, score_as_key)
    
    def classify(self, pil_image, stride, batch_size, score_as_key=False):
        """ Classifies a PIL image, returning a map with class names and
            corresponding scores.
            
            Parameters
            ----------
                pil_image: PIL image
                    Image to classify
                stride: int
                    The CNN is applied patch-wise; this parameter
                    corresponds to the offset between two patches
                batch_size: int
                    Number of patches which can be processed at the same
                    time by the hardware. If no GPU is used, then a
                    value of 1 is fine.
                score_as_key: bool
                    Use scores, instead of class names, as key for the
                    result map.
            
            Returns
            -------
                A map between class names and scores, or scores and
                class names, depending on whether score_as_key is true
                or false.
        """
        
        if pil_image.size[0]>1000:
            pil_image = pil_image.resize((1000, round(pil_image.size[1]*1000.0/pil_image.size[0])), Image.BILINEAR)
        tensorize = transforms.ToTensor()
        was_training = self.network.training
        self.network.eval()
        with torch.no_grad():
            score = 0
            processed_samples = 0
            batch = []
            for x in range(0, pil_image.size[0], stride):
                for y in range(0, pil_image.size[1], stride):
                    crop = tensorize(pil_image.crop((x, y, x+224, y+224)))
                    batch.append(crop)
                    if len(batch) >= batch_size:
                        tensors = torch.stack(batch).to(self.dev)
                        out = self.network(tensors)
                        score += out.sum(0)
                        processed_samples += len(batch)
                        batch = []
            if batch:
                tensors = torch.stack(batch).to(self.dev)
                out = self.network(tensors)
                score += out.sum(0)
                processed_samples += len(batch)
                batch = []
        if was_training:
            self.network.train()
        score /= processed_samples
        res = {}
        for cl in self.classMap.cl2id:
            cid = self.classMap.cl2id[cl]
            if cid == -1:
                continue
            res[cl] = score[cid].item()
        if score_as_key:
            res = {s: c for c, s in res.items()}
        return res
    
    def __repr__(self):
        """ returns a string description of the instance """
        
        format_string = self.__class__.__name__ + '('
        format_string += '\n ClassMap: %s' % self.classMap
        format_string += '\n Network:'
        if self.network is None:
            format_string += '\n  None'
        else:
            format_string += '\n%s\nEnd of network\n' % self.network
        return format_string+'\n)'
    



import torch
from torch.nn.modules import Module

class DontCareLoss(Module):
    def __init__(self, dimensionality):
        super(DontCareLoss, self).__init__()
        
    def forward(self, input, target, dont_care):
        loss = 0
        for n in range(input.size(0)):
            t  = target[n]
            dc = dont_care[n]
            for i in range(input[n].size(0)):
                if i==t:
                    loss += (1 - input[n][i])**2
                elif not i in dc:
                    loss += input[n][i]**2
        return loss



"""
OCR-D conformant command line interface
"""
import click
from ocrd.decorators import ocrd_cli_options, ocrd_cli_wrap_processor

from ..processor import TypegroupsClassifierProcessor

@click.command()
@ocrd_cli_options
def cli(*args, **kwargs):
    """
    Classify typegroups
    """
    return ocrd_cli_wrap_processor(TypegroupsClassifierProcessor, *args, **kwargs)



"""
    Model definition adapted from: https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py
    The differences are:
        - insertion of variational convolutions in the layers
        - size adaptation by replacing the last avg pooling by an adaptive
          one
    - methods turning this class into an auto-encoder
"""
import logging
import math

import torch.nn as nn
from ocrd_typegroups_classifier.network.var_conv2d import VarConv2d

def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)

def varConv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return VarConv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


class _VariationalBasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(_VariationalBasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.relu = nn.ReLU()
        self.conv2 = varConv3x3(planes, planes)
        self.downsample = downsample
        self.stride = stride
        
        self.varloss = 0

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.relu(out)

        out, self.varloss = self.conv2(out)
        
        if self.downsample is not None:
            residual = self.downsample(x)

        out = out + residual
        out = self.relu(out)

        return out


class _VariationalBottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(_VariationalBottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = VarConv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU()
        self.downsample = downsample
        self.stride = stride
        
        self.varloss = 0

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out, self.varloss = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out = out + residual
        out = self.relu(out)

        return out


class _VRAEC(nn.Module):

    def __init__(self, block, layers, layer_size=64, output_channels=256):
        self.inplanes = 64
        super(_VRAEC, self).__init__()
        
        if isinstance(layer_size, int):
            layer_size = (layer_size, layer_size, layer_size, layer_size)

        self.expected_input_size = (224, 224)
        self.layer_size = layer_size
        self.block = block

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.relu = nn.ReLU()
        self.tanh = nn.Tanh()
        self.maxpool = nn.Conv2d(64, 64, kernel_size=3, stride=2, padding=1)
        self.layer1, self.vl1 = self._make_layer(block, layer_size[0], layers[0])
        self.layer2, self.vl2 = self._make_layer(block, layer_size[1], layers[1], stride=2)
        self.layer3, self.vl3 = self._make_layer(block, layer_size[2], layers[2], stride=2)
        self.layer4, self.vl4 = self._make_layer(block, layer_size[3], layers[3], stride=2)
        #self.avgpool = nn.AvgPool2d(7, stride=1)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Linear(layer_size[3] * block.expansion, output_channels)

        
        self.deconv1    = nn.ConvTranspose2d(64, 3, kernel_size=8, stride=2, padding=3)
        self.unpool     = nn.ConvTranspose2d(64, 64, kernel_size=2, stride=2, padding=0)
        self.unlayer1   = nn.ConvTranspose2d(block.expansion * layer_size[0], 64, kernel_size=3, padding=1)
        self.unlayer2   = nn.ConvTranspose2d(block.expansion * layer_size[1], block.expansion * layer_size[0], kernel_size=4, stride=2, padding=1)
        self.unlayer3   = nn.ConvTranspose2d(block.expansion * layer_size[2], block.expansion * layer_size[1], kernel_size=4, stride=2, padding=1)
        self.unlayer4   = nn.ConvTranspose2d(block.expansion * layer_size[3], block.expansion * layer_size[2], kernel_size=4, stride=2, padding=1)
        
        
        self.ae_layers = {
            0: (self.conv1, self.deconv1),
            1: (self.unpool, ),
            2: (self.layer1, self.unlayer1),
            3: (self.layer2, self.unlayer2),
            4: (self.layer3, self.unlayer3),
            5: (self.layer4, self.unlayer4),
            6: (self.avgpool,),
            7: (self.fc,)
        }
        
        self.var_layers = {
            0: None,
            1: None,
            2: self.vl1,
            3: self.vl2,
            4: self.vl3,
            5: self.vl4,
            6: None # sentinel
        }

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        layers[0].conv2.is_variational = False
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            b = block(self.inplanes, planes)
            b.conv2.is_variational = False
            layers.append(b)

        return (nn.Sequential(*layers), b)

    def forward(self, x):
        # input size: torch.Size([2, 3, 224, 224])
        
        x = self.conv1(x)
        x = self.relu(x)
        # after convo1: torch.Size([2, 64, 112, 112])
        
        x = self.maxpool(x)
        # after maxpool: torch.Size([2, 64, 56, 56])

        x = self.layer1(x)
        vl = self.vl1.varloss
        # after layer 1: torch.Size([2, 96, 56, 56])
        
        x = self.layer2(x)
        vl = vl + self.vl2.varloss
        # after layer 2: torch.Size([2, 96, 28, 28])
        
        x = self.layer3(x)
        vl = vl + self.vl3.varloss
        # after layer 3: torch.Size([2, 96, 14, 14])
        
        x = self.layer4(x)
        vl = vl + self.vl4.varloss
        # after layer 4: torch.Size([2, 96, 7, 7])
        
        ap = self.avgpool(x)
        # after avgpool: torch.Size([2, 96, 1, 1])
        
        x = ap.view(ap.size(0), -1)
        x = self.fc(x)

        return x, vl, ap.detach()
    
    def encode(self, x, nb_layers):
        px = x
        x = self.relu(self.conv1(x))
        
        poolpos = None
        if nb_layers>=1:
            px = x
            x = self.maxpool(x)
        
        layers = {
            2: (self.layer1, 'l1', self.vl1),
            3: (self.layer2, 'l2', self.vl2),
            4: (self.layer3, 'l3', self.vl3),
            5: (self.layer4, 'l4', self.vl4),
            6: None
        }
        vl = 0
        for l in range(2, nb_layers+1):
            if layers[l] is None:
                break
            px = x
            x = layers[l][0](x)
            vl = vl + layers[l][2].varloss
        
        return x, px, vl
    
    def set_variational(self, lnum, status):
        layers = {
            2: self.vl1,
            3: self.vl2,
            4: self.vl3,
            5: self.vl4
        }
        if lnum in layers:
            l = layers[lnum]
            l.conv2.is_variational = status
        
    
    def decode(self, x, layers):
        if 5 in layers:
            x = self.relu(self.unlayer4(x))
        
        if 4 in layers:
            x = self.relu(self.unlayer3(x))
        
        if 3 in layers:
            x = self.relu(self.unlayer2(x))
        
        if 2 in layers:
            x = self.relu(self.unlayer1(x))
        
        if 1 in layers:
            x = self.unpool(x)
        
        if 0 in layers:
            x = self.tanh(self.deconv1(x))
        return x
    
    def train_ae(self, x, optimizer, loss_function, layer_num=6):
        enc, penc, vl = self.encode(x, layer_num)
        dec = self.decode(enc, (layer_num,))
        optimizer.zero_grad()
        loss = loss_function(dec, penc.detach())
        
        if not vl is 0:
            loss = loss + vl / (vl.item()/loss.item())
        loss.backward()
        optimizer.step()
        return loss.item()
        
    
    def finetune(self, x, optimizer, loss_function, layer_num=6):
        layers = range(layer_num+1)
        enc, _, vl = self.encode(x, layer_num)
        dec = self.decode(enc, layers)
        optimizer.zero_grad()
        loss = loss_function(dec, x)
        if not vl is 0:
            loss = loss + vl / (vl.item()/loss.item())
        loss.backward()
        optimizer.step()
        return loss.item()

    def select_parameters(self, layers=range(8)):
        res = list()
        
        for l in layers:
            for layer in self.ae_layers[l]:
                for p in layer.parameters():
                    res.append(p)
        return res


def vraec18(pretrained=False, **kwargs):
    """Constructs a _ResAE-18 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = _VRAEC(_VariationalBasicBlock, [2, 2, 2, 2], **kwargs)
    if pretrained:
        try:
            model.load_state_dict(model_zoo.load_url(model_urls['resnet18']), strict=False)
        except Exception as exp:
            logging.warning(exp)
    return model


def vraec34(pretrained=False, **kwargs):
    """Constructs a _ResAE-34 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = _VRAEC(_VariationalBasicBlock, [3, 4, 6, 3], **kwargs)
    if pretrained:
        try:
            model.load_state_dict(model_zoo.load_url(model_urls['resnet34']), strict=False)
        except Exception as exp:
            logging.warning(exp)
    return model


def vraec50(pretrained=False, **kwargs):
    """Constructs a _ResAE-50 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = _VRAEC(_VariationalBottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        try:
            model.load_state_dict(model_zoo.load_url(model_urls['resnet50']), strict=False)
        except Exception as exp:
            logging.warning(exp)
    return model


def vraec101(pretrained=False, **kwargs):
    """Constructs a _ResAE-101 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = _VRAEC(_VariationalBottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        try:
            model.load_state_dict(model_zoo.load_url(model_urls['resnet101']), strict=False)
        except Exception as exp:
            logging.warning(exp)
    return model


def vraec152(pretrained=False, **kwargs):
    """Constructs a _ResAE-152 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = _VRAEC(_VariationalBottleneck, [3, 8, 36, 3], **kwargs)
    if pretrained:
        try:
            model.load_state_dict(model_zoo.load_url(model_urls['resnet152']), strict=False)
        except Exception as exp:
            logging.warning(exp)
    return model



class IndexRemap:
    """ Utility class for remapping class indices.
    
        Attributes
        ----------
        id2id: dictionary
            map from source to target index
    """
    
    def __init__(self, id2id):
        """ 
            Parameters
            ----------
            
            id2id: dictionary int to int
                map from source to target index
        """
        self.id2id = id2id
    
    def __call__(self, n):
        """ Remaps an index, returns -1 if the input index is not known
        
        """
        if not n in self.id2id:
            return -1
        return self.id2id[n]
    
    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for k in self.id2id:
            format_string += '\n %d:%d' % (k, self.id2id[k])
        return format_string+'\n)'


class ClassMap:
    """ Class wrapping type group information and a classifier.
    
        Attributes
        ----------
        
        cl2id: dictionary string to int
            Maps a class name to a class number
        id2cl: dictionary int to string
            Maps a class number to a class name
    """
    
    def __init__(self, basemap):
        """ Constructor of the class.
        
            Parameters
            ----------
            
            basemap: map string to int
                Maps names to IDs with regard to the network outputs;
                note that several names can point to the same ID, but
                the inverse is not possible.
        """
        
        self.cl2id = {}
        self.id2cl = {}
        for c in basemap:
            self.cl2id[c] = basemap[c]
            self.id2cl[basemap[c]] = c
    
    def forget_class(self, target):
        del self.id2cl[self.cl2id[target]]
        del self.cl2id[target]
    
    def translate(self, dictionary):
        for name in self.cl2id:
            if not name in dictionary:
                continue
            n = self.cl2id[name]
            del self.cl2id[name]
            self.cl2id[dictionary[name]] = n
            self.id2cl[n] = dictionary[name]
       
    def get_target_transform(self, dataset_classes):
        """ Creates a transform from a map (class name to id) to the
            set of IDs used by this class map. Unmatched classes are
            mapped to -1.
            
            This method is useful for producing target transforms for
            PyTorch ImageFolder. Proceed as follows:
                imf = ImageFolder('/path')
                imf.target_transform = cm.get_target_transform(imf.class_to_idx)
            where cm is a ClassMap instance.
        """
        
        idmap = {}
        for c in dataset_classes:
            if not c in self.cl2id:
                idmap[dataset_classes[c]] = -1
            else:
                idmap[dataset_classes[c]] = self.cl2id[c]
        return IndexRemap(idmap)
    
    def __repr__(self):
        return '%s' % self.cl2id



from io import BytesIO
from PIL import Image
from random import randint

class QLoss(object):
    """ Quality loss-based data augmentation method compatible with PyTorch.
    
        Applied a random JPEG compression to the input PIL image. By default,
        the quality is between 1 and 100.
    """
    
    def __init__(self, min_q=1, max_q=100):
        """ 
            Parameters
            ----------
            
            min_q: int
                minimum compression quality (automatically floored to 1)
            max_q: int
                maximum compression quality (ceiled to 100) - note that
                this value must not be lower than min_q.
        """
        
        self.min_q = max(min_q, 1)
        self.max_q = min(max_q, 100)
        assert self.min_q <= self.max_q

    def __call__(self, img):
        """ Returns a copy of an image with JPG degradations
            
            Parameters
            ----------
            
            img: PIL.Image
            
            Returns
            -------
                res: PIL.Image
        """
        
        temp = BytesIO()
        q = randint(self.min_q, self.max_q)
        img.save(temp, format='jpeg', quality=q)
        res = Image.open(temp)
        return res
    
    def __repr__(self):
        """ Returns a string description of the class """
        return self.__class__.__name__ + '(min_q={}, max_q={})'.format(self.min_q, self.max_q)



from json import loads

from pkg_resources import resource_string

classes = {
    0: "Antiqua",
    1: "Bastarda",
    6: "Rotunda",
    7: "Textura"
}

OCRD_TOOL = loads(resource_string(__name__, 'ocrd-tool.json'))



from PIL import Image
from skimage import filters
from random import randint
import numpy as np

class Sauvola(object):
    """ Sauvola binarization data augmentation method compatible with PyTorch.
    """
    
    def __init__(self, min_r=2, max_r=10):
        """ 
            Parameters
            ----------
            
            min_r: int
                minimum radius 
            max_q: int
                maximum compression quality (ceiled to 100) - note that
                this value must not be lower than min_q.
        """
        
        self.min_r = max(min_r, 1)
        self.max_r = min(max_r, 50)
        assert self.min_r <= self.max_r

    def __call__(self, img):
        """ Returns a binarized version of the image
            
            Parameters
            ----------
            
            img: PIL.Image
            
            Returns
            -------
                res: PIL.Image
        """
        
        r = 1+2*randint(self.min_r, self.max_r)
        arr = np.array(img)
        t = filters.threshold_sauvola(arr, window_size=r)
        arr[arr<t]  =   0
        arr[arr>=t] = 255
        return Image.fromarray(arr)
    
    def __repr__(self):
        """ Returns a string description of the class """
        return self.__class__.__name__ + '(min_r={}, max_r={})'.format(self.min_r, self.max_r)


class Otsu(object):
    """ Sauvola binarization data augmentation method compatible with PyTorch.
    """
    
    def __init__(self):
        """ 
            Parameters
            ----------
            None
        """

    def __call__(self, img):
        """ Returns a binarized version of the image
            
            Parameters
            ----------
            
            img: PIL.Image
            
            Returns
            -------
                res: PIL.Image
        """
        
        arr = np.array(img.convert('LA'))
        try:
                t = filters.threshold_otsu(arr)
        except:
                return img # probably single-value image
        arr[arr<t]  =   0
        arr[arr>=t] = 255
        return Image.fromarray(arr).convert('RGB')
    
    def __repr__(self):
        """ Returns a string description of the class """
        return self.__class__.__name__ + '()'



"""
standalone command line interface
"""
import sys

from PIL import Image

from ..typegroups_classifier import TypegroupsClassifier

def cli():
    """
    Run on sys.args
    """
    def show_help(retval=1):
        print('Syntax: %s network-file image-file [stride]' % sys.argv[0])
        sys.exit(retval)
    if any([x in ('-h', '--help') for x in sys.argv]):
        show_help(0)
    if len(sys.argv) < 3 or len(sys.argv) > 4:
        show_help(1)
    network_file = sys.argv[1]
    image_file = Image.open(sys.argv[2])
    stride = int(sys.argv[3]) if len(sys.argv) > 3 else 96
    classifier = TypegroupsClassifier(network_file, stride)
    result = classifier.run(image_file)
    print(result)



"""
Wrap TypegroupsClassifier as an ocrd.Processor
"""
from pkg_resources import resource_filename

from ocrd import Processor
from ocrd_utils import (
    getLogger,
    make_file_id,
    assert_file_grp_cardinality,
    MIMETYPE_PAGE
)
from ocrd_utils import getLogger, make_file_id, MIMETYPE_PAGE
from ocrd_models.ocrd_page import (
    to_xml,
    TextStyleType
)
from ocrd_modelfactory import page_from_file

from .typegroups_classifier import TypegroupsClassifier
from .constants import OCRD_TOOL


class TypegroupsClassifierProcessor(Processor):

    def __init__(self, *args, **kwargs):
        kwargs['ocrd_tool'] = OCRD_TOOL['tools']['ocrd-typegroups-classifier']
        kwargs['version'] = OCRD_TOOL['version']
        super(TypegroupsClassifierProcessor, self).__init__(*args, **kwargs)

    def process(self):
        """Classify historic script in pages and annotate as font style.

        Open and deserialize PAGE input files and their respective images.
        Then for each page, retrieve the raw image and feed it to the font
        classifier. 

        Post-process detections by filtering classes and thresholding scores.
        Annotate the good predictions by name and score as a comma-separated
        list under ``/PcGts/Page/TextStyle/@fontFamily``, if any.

        Produce a new PAGE output file by serialising the resulting hierarchy.
        """
        log = getLogger('ocrd_typegroups_classifier')
        assert_file_grp_cardinality(self.input_file_grp, 1)
        assert_file_grp_cardinality(self.output_file_grp, 1)
        if 'network' not in self.parameter:
            self.parameter['network'] = resource_filename(__name__, 'models/densenet121.tgc')
        network_file = self.resolve_resource(self.parameter['network'])
        stride = self.parameter['stride']
        classifier = TypegroupsClassifier.load(network_file)

        ignore_type = ('Adornment', 'Book covers and other irrelevant data',
                       'Empty Pages', 'Woodcuts - Engravings')

        for n, input_file in enumerate(self.input_files):
            page_id = input_file.pageId or input_file.ID
            log.info('Processing: %d / %s', n, page_id)
            pcgts = page_from_file(self.workspace.download_file(input_file))
            self.add_metadata(pcgts)
            page = pcgts.get_Page()
            pil_image, _, image_info = self.workspace.image_from_page(
                # prefer raw image (to meet expectation of the models, which
                # have been trained on RGB images with both geometry and color
                # transform random augmentation)
                # maybe even: dewarped,deskewed ?
                page, page_id, feature_filter='binarized,normalized,grayscale_normalized,despeckled')
            # todo: use image_info.resolution
            result = classifier.run(pil_image, stride)
            score_sum = 0
            for typegroup in classifier.classMap.cl2id:
                if not typegroup in ignore_type:
                    score_sum += max(0, result[typegroup])

            script_highscore = 0
            noise_highscore = 0
            result_map = {}
            output = ''
            for typegroup in classifier.classMap.cl2id:
                score = result[typegroup]
                if typegroup in ignore_type:
                    noise_highscore = max(noise_highscore, score)
                else:
                    script_highscore = max(script_highscore, score)
                    normalised_score = max(0, score / score_sum)
                    result_map[normalised_score] = typegroup
            if noise_highscore > script_highscore:
                page.set_primaryScript(None)
                log.warning(
                    'Detected only noise on page %s. noise_highscore=%s > script_highscore=%s',
                    page_id, noise_highscore, script_highscore)
            else:
                for k in sorted(result_map, reverse=True):
                    intk = round(100*k)
                    if intk<=0:
                        continue
                    if output != '':
                        output = '%s, ' % output
                    output = '%s%s:%d' % (output, result_map[k], intk)
                log.debug('Detected %s' % output)
                textStyle = page.get_TextStyle()
                if not textStyle:
                    textStyle = TextStyleType()
                    page.set_TextStyle(textStyle)
                textStyle.set_fontFamily(output)
                file_id = make_file_id(input_file, self.output_file_grp)
                pcgts.set_pcGtsId(file_id)
                self.workspace.add_file(
                    ID=file_id,
                    file_grp=self.output_file_grp,
                    pageId=input_file.pageId,
                    mimetype=MIMETYPE_PAGE,
                    local_filename="%s/%s.xml" % (self.output_file_grp, file_id),
                    content=to_xml(pcgts)
                )



# -*- coding: utf-8 -*-
import codecs
import json

from setuptools import setup, find_packages

with codecs.open('README.md', encoding='utf-8') as f:
    README = f.read()
with open('./ocrd-tool.json', 'r') as f:
    version = json.load(f)['version']

setup(
    name='ocrd_typegroups_classifier',
    version=version,
    description='Typegroups classifier for OCR',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Matthias Seuret, Konstantin Baierer',
    author_email='seuretm@users.noreply.github.com',
    url='https://github.com/OCR-D/ocrd_typegroups_classifier',
    license='Apache License 2.0',
    packages=find_packages(exclude=('tests', 'docs')),
    include_package_data=True,
    install_requires=open('requirements.txt').read().split('\n'),
    package_data={
        '': ['*.json', '*.tgc'],
    },
    entry_points={
        'console_scripts': [
            'typegroups-classifier=ocrd_typegroups_classifier.cli.simple:cli',
            'ocrd-typegroups-classifier=ocrd_typegroups_classifier.cli.ocrd_cli:cli',
        ]
    },
)



# coding=utf-8
import math
import torch
from torch import nn
from torch.nn.parameter import Parameter
from torch.nn import Module
from torch.nn import Conv2d

# This class use some of the code from
# https://github.com/pytorch/examples/blob/master/vae/main.py

class VarConv2d(Module):
    """ Variational convolution 2D module for PyTorch.
    
    This class is based on the Variational Auto-Encoder; more details are
    given in the code comments.
    
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, is_variational=True):
        """ Constructor of the class, following the same syntax as the
            PyTorch Conv2D class, with an optional is_variational
            parameter.
        
            Note that by default it is variational. You can switch on
            and off this behavior by modifying the value of the attribute
            is_variational.
        """
        
        super(VarConv2d, self).__init__()
        self.in_channels  = in_channels
        self.out_channels = out_channels
        self.bias = bias
        self.kernel_size = kernel_size
        self.padding = padding
        self.stride = stride
        self.is_variational = is_variational
        
        self.mu_layer      = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation, groups=groups, bias=bias)
        self.logvar_layer  = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation, groups=groups, bias=bias)
        

    def forward(self, x):
        """ Forward step of the variational convolution
            
            Parameters
            ----------
                x: PyTorch tensor / batch
                    Data to process
            
            Returns
            -------
                (output, variational loss)
                    A pair composed of the output tensor, and the variational
                    loss. If the layer is not being trained or if its
                    is_variational attribute is false, then the variational
                    loss is 0.
        """
        mu = self.mu_layer(x)
        if self.training and self.is_variational:
            logvar = self.logvar_layer(x)
            std = torch.exp(0.5*logvar)
            eps = torch.randn_like(std)
            mu = eps.mul(std).add_(mu)
            varloss = self.kld(mu, logvar)
        else:
            varloss = 0
        return (mu, varloss)
    
    def kld(self, mu, logvar):
        """ Computes the Kullback-Leibler Divergence, which is used as
        an extra-loss when training the variational layer """
        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        return -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
    
    def extra_repr(self):
        """ Returns a string description of the instance """
        return 'out_channels={out_channels}, kernel_size={kernel_size}, ' \
            'padding={padding}, stride={stride}, bias={bias}, is_variational={is_variational}'.format(
                **self.__dict__)
